"""The exceptions used in Walytis_Beta API.

See `walytis_beta_tools.exceptions`
"""

from walytis_beta_tools.exceptions import(
    WalytisReplyDecodeError,
    WalytisBugError,
    NoSuchBlockchainError,
    BlockchainAlreadyExistsError,
    NoSuchInvitationError,
    JoinFailureError,
    BlockNotFoundError,
    BlockCreationError,
    InvalidBlockIdError,
    BlockIntegrityError,
    NotSupposedToHappenError,
)