"""The versions of the the various Walytis components.

See `walytis_beta_tools.versions`
"""
from walytis_beta_tools.versions import (
    WALYTIS_BETA_PROTOCOL_VERSION,
    WALYTIS_BETA_API_PROTOCOL_VERSION,
    WALYTIS_BETA_API_VERSION,
    WALYTIS_BETA_CORE_VERSION,
)

